@@ -319,6 +319,15 @@ public class BundleRevisionImpl implements BundleRevision, Resource
         }
     }
 
+    synchronized void disposeContentPath()
+    {
+        for (int i = 0; (m_contentPath != null) && (i < m_contentPath.size()); i++)
+        {
+            m_contentPath.get(i).close();
+        }
+        m_contentPath = null;
+    }
+
     public void setProtectionDomain(ProtectionDomain pd)
     {
         m_protectionDomain = pd;
@@ -338,18 +347,19 @@ public class BundleRevisionImpl implements BundleRevision, Resource
         return m_content;
     }
 
-    void resetContent(Content content)
+    synchronized void resetContent(Content content)
     {
         m_content = content;
     }
 
     List<Content> getContentPath()
     {
-        if (m_contentPath == null)
+        List<Content> contentPath = m_contentPath;
+        if (contentPath == null)
         {
             try
             {
-                m_contentPath = initializeContentPath();
+                contentPath = initializeContentPath();
             }
             catch (Exception ex)
             {
@@ -357,7 +367,7 @@ public class BundleRevisionImpl implements BundleRevision, Resource
                     m_bundle, Logger.LOG_ERROR, "Unable to get module class path.", ex);
             }
         }
-        return m_contentPath;
+        return contentPath;
     }
 
     private synchronized List<Content> initializeContentPath() throws Exception
@@ -388,7 +398,7 @@ public class BundleRevisionImpl implements BundleRevision, Resource
                     fragments.get(i), fragmentContents.get(i), contentList, false);
             }
         }
-        return contentList;
+        return m_contentPath = contentList;
     }
 
     private List calculateContentPath(
@@ -505,6 +515,10 @@ public class BundleRevisionImpl implements BundleRevision, Resource
         {
             if (contentPath.get(i).hasEntry(name))
             {
+                if (!name.endsWith("/") && contentPath.get(i).isDirectory(name))
+                {
+                    name += "/";
+                }
                 url = createURL(i + 1, name);
             }
         }
@@ -543,6 +557,10 @@ public class BundleRevisionImpl implements BundleRevision, Resource
             {
                 if (contentPath.get(i).hasEntry(name))
                 {
+                    if (!name.endsWith("/") && contentPath.get(i).isDirectory(name))
+                    {
+                        name += "/";
+                    }
                     // Use the class path index + 1 for creating the path so
                     // that we can differentiate between module content URLs
                     // (where the path will start with 0) and module class
@@ -579,6 +597,10 @@ public class BundleRevisionImpl implements BundleRevision, Resource
             // Check the module content.
             if (getContent().hasEntry(name))
             {
+                if (!name.endsWith("/") && getContent().isDirectory(name))
+                {
+                    name += "/";
+                }
                 // Module content URLs start with 0, whereas module
                 // class path URLs start with the index into the class
                 // path + 1.
@@ -616,6 +638,25 @@ public class BundleRevisionImpl implements BundleRevision, Resource
         return getContentPath().get(index - 1).getEntryAsStream(urlPath);
     }
 
+
+    public long getContentTime(int index, String urlPath)
+    {
+        if (urlPath.startsWith("/"))
+        {
+            urlPath = urlPath.substring(1);
+        }
+        Content content;
+        if (index == 0)
+        {
+            content = getContent();
+        }
+        else {
+            content = getContentPath().get(index - 1);
+        }
+        long result = content.getContentTime(urlPath);
+        return result > 0 ? result : m_bundle.getLastModified();
+    }
+
     public URL getLocalURL(int index, String urlPath)
     {
         if (urlPath.startsWith("/"))
@@ -670,11 +711,7 @@ public class BundleRevisionImpl implements BundleRevision, Resource
         }
         m_content.close();
         m_content = null;
-        for (int i = 0; (m_contentPath != null) && (i < m_contentPath.size()); i++)
-        {
-            m_contentPath.get(i).close();
-        }
-        m_contentPath = null;
+        disposeContentPath();
     }
 
     @Override
